var _ = require( "lodash" ),
    couchbase = require( "couchbase" ),
    machina = require( "machina" )( _ ),
    createFsm = function( options ) {

        var magicMethods = [
            "touch",
            "get",
            "set",
            "remove",
            "replace",
            "add",
            "append",
            "prepend",
            "view",
            "incr",
            "decr",
            "observe",
            "endure",
            "setDesignDoc",
            "getDesignDoc",
            "deleteDesignDoc",
            "getVersion",
            "strError"
        ],
            fsm = {
                client: null,

                connect: function() {
                    var self = this,
                        allowed_keys = ['debug', 'user', 'username', 'password', 'pass', 'passw', 'hosts', 'host', 'hostname', 'bucket'],
                        connect_options = _.omit( options, function(val, key) {
                            return ! _.contains( allowed_keys, key );
                        });

                    couchbase.connect( connect_options, function( err, client ) {
                        if ( err ) {
                            self.emit( 'fail' );
                            return;
                        }
                        self.client = client;
                        self.transition( 'connected' );
                    });
                },

                initialState: 'disconnected',

                states: {
                    disconnected: {

                        _onEnter: function() {
                            this.connect();
                        },

                        "*": function() {
                            this.deferUntilTransition( 'connected' );
                        }

                    },

                    connected: {
                        "*": function( method, args ) {
                            this.client[ method ].apply( this.client, args );
                        }
                    }
                }

            };

        _.each( magicMethods, function( method ) {
            fsm[ method ] = function() {
                var args = Array.prototype.slice.call( arguments );
                this.handle.call( this, method, args );
            };
        });

        var ClientFsm = new machina.Fsm( fsm );

        return ClientFsm;
    },
    Client = {
        registry: {},
        subscriptions: [],
        bus: null,
        createClient: function( name, config ) {
            var client = createFsm( config );
            Client.registry[ name ] = client;
            return client;
        },
        getClient: function( name ) {
            return Client.registry[ name ];
        },
        methods: {
            send: function( topic, envelope, res ) {

                if ( envelope.replyTo ) {
                    process.nextTick(function() {
                        var response = _.omit( envelope, 'channel', 'topic', 'data', 'timeStamp' );
                        _.merge( response, {
                            channel: envelope.replyTo.channel,
                            topic: envelope.replyTo.topic,
                            data: res
                        });
                        return Client.bus.publish( response );
                    });
                    
                }

                return;
            },
            init: function( msg, envelope ) {
                var config = envelope.db;

                if ( ! Client.registry[ envelope.resource ] ) {
                    var client = Client.createClient( envelope.resource, config );
                    // Create design document
                    // @TODO Figure out how to handle bucket initialization of views
                    var ddoc = {
                         "views": {
                             "all": {
                                 "map": "function( doc, meta ){ emit( meta.id, doc ); }"
                             }
                         }
                     };

                     client.setDesignDoc( config.bucket, ddoc, function( err, data ) {

                     });
                }
            },
            get: function( msg, envelope ) {
                var data = msg,
                    client = Client.getClient( envelope.resource ),
                    response = {};

                if ( _.isArray( data.id ) ) {
                    client.handle( 'get', [ data.id, null, function( err, docs, metas ) {
                        if ( err ) {
                            if ( err.code === 13 ) {
                                // Key not found
                                response.err = {
                                    statusCode: 404,
                                    message: 'not found'
                                };
                            } else {
                                response.err = err;
                            }

                            return Client.methods.send( 'get', envelope, response );
                        }

                        for( var i = 0, j = docs.length; i < j; i++ ) {
                            docs[ i ].id = metas[ i ].id;
                        }
                        response.data = docs;
                        
                        return Client.methods.send( 'get', envelope, response );

                    } ] );
                } else {
                    client.handle( 'get', [ data.id, function( err, result, meta ) {
                        if ( err ) {
                            if ( err.code === 13 ) {
                                // Key not found
                                response.err = {
                                    statusCode: 404,
                                    message: 'not found'
                                };
                            } else {
                                response.err = err;
                            }

                            return Client.methods.send( 'get', envelope, response );
                        }
                         
                        result.id = meta.id;
                        response.data = result;
                        
                        return Client.methods.send( 'get', envelope, response );

                    }] );
                }

            },
            all: function( msg, envelope ) {
                var data = msg,
                    client = Client.getClient( envelope.resource ),
                    response = {};

                client.handle( 'view', [ envelope.db.bucket, "all", data.options, function( err, results ) {
                    if ( err ) {
                        response.err = err;
                        return Client.methods.send( 'all', envelope, response );
                    }

                    var list = [],
                        row = null;

                    for( var i = 0, j = results.length; i < j; i++ ) {
                        row = results[ i ].value;
                        row.id = results[ i ].id;
                        list.push( row );
                    }

                    response.data = list;

                    return Client.methods.send( 'all', envelope, response );
                }] );
            },
            find: function( msg, envelope ) {
                var data = msg,
                    client = Client.getClient( envelope.resource ),
                    response = {},
                    view_name,
                    view;

                if ( data.options.path ) {
                    view_name = data.options.path;
                } else {
                    var view_key = _.keys( msg.keys ).sort().join("_"),
                        views = envelope.db.views;

                    view_name = views[ view_key ];
                }

                if ( ! view_name ) {
                    response.err = {
                        statusCode: 404,
                        message: 'View not found'
                    };
                    return Client.methods.send( 'find', envelope, response );
                }

                view = view_name.split( "/" );

                data.options = data.options || {};

                // This is a temporary workaround for getting keys
                var keys = _.values( msg.keys );
                if ( keys.length === 1 ) {
                    data.options.key = keys[0];
                } else {
                    data.options.keys = keys;
                }

                client.handle( 'view', [ view[0], view[1], data.options, function( err, results ) {
                    if ( err ) {
                        response.err = err;
                        return Client.methods.send( 'find', envelope, response );
                    }

                    var list = [],
                        row = null;

                    for( var i = 0, j = results.length; i < j; i++ ) {
                        row = results[ i ].value;
                        row.id = results[ i ].id;
                        row.key = results[ i ].key;
                        list.push( row );
                    }
                    response.data = list;

                    return Client.methods.send( 'find', envelope, response );
                } ]);
            },
            save: function( msg, envelope ) {
                var data = msg,
                    client = Client.getClient( envelope.resource ),
                    response = {};

                client.handle( 'set', [ data.id, data.data, function( err, result, meta ) {
                    if ( err ) {
                        response.err = err;
                        return Client.methods.send( 'save', envelope, response );
                    }

                    response.data = result;
                    return Client.methods.send( 'save', envelope, response );
                }] );
            },
            destroy: function( msg, envelope ) {
                var data = msg,
                    client = Client.getClient( envelope.resource ),
                    response = {};

                client.handle( 'remove', [ data.id, function( err, result ) {
                    if ( err ) {
                        response.err = err;
                        return Client.methods.send( 'destroy', envelope, response );
                    }

                    response.data = result;
                    return Client.methods.send( 'destroy', envelope, response );
                } ] );
            }
        },
        listen: function( options ) {
            var bus = Client.bus = options.bus,
                listenChannel = options.config.dataListenChannel;

            var sub1 = bus.subscribe({
                channel: listenChannel,
                topic: "init",
                callback: Client.methods.init
            });
            
            var sub2 = bus.subscribe({
                channel: listenChannel,
                topic: "get",
                callback: Client.methods.get
            });

            var sub3 = bus.subscribe({
                channel: listenChannel,
                topic: "all",
                callback: Client.methods.all
            });

            var sub4 = bus.subscribe({
                channel: listenChannel,
                topic: "find",
                callback: Client.methods.find
            });

            var sub5 = bus.subscribe({
                channel: listenChannel,
                topic: "save",
                callback: Client.methods.save
            });

            var sub6 = bus.subscribe({
                channel: listenChannel,
                topic: "destroy",
                callback: Client.methods.destroy
            });

            Client.subscriptions.push( sub1 );
            Client.subscriptions.push( sub2 );
            Client.subscriptions.push( sub3 );
            Client.subscriptions.push( sub4 );
            Client.subscriptions.push( sub5 );
            Client.subscriptions.push( sub6 );
        },
        off: function() {
            _.each( Client.subscriptions, function( sub ) {
                if( sub.unsubscribe ) {
                    sub.unsubscribe();
                }
            });

            Client.subscriptions = [];
        }
    };


module.exports = Client;