var _ = require( "lodash" ),
    inflect = require( "inflect" ),
    riveter = require( "riveter" )( _ ),
    postal = require( "postal" )( _ ),
    clientFsm = require( './clientfsm.js' ),
    uuid = require( "node-uuid" ),
    noop = function(){};

module.exports = function( resource ) {

    var ResourceFactory = function( name ){
        this.name = inflect.singularize( name.toLowerCase() );
        this.name_plural = inflect.pluralize( this.name );
        this._config = {};
        this._embedded = false;
        this._key_prefix = this.name + "::";
        this._beforeHooks = {};
        this._afterHooks = {};
        this._schema = {};
        this._instanceMethods = {};

        var self = this;

        this.__defineSetter__( 'embedded', function( value ) {
            if ( value === true ) {
                _.merge( self, resource.mixins.embedded.factory );
            } else {
                _.merge( self, resource.mixins["default"].factory );
            }
            
            this._embedded = value;
        });

        this.__defineGetter__( 'embedded', function() {
            return this._embedded;
        });

    };

    ResourceFactory.prototype.db = function( topic, data, done ) {
        var env = {
            channel: resource.config.dataRequestChannel,
            topic: topic,
            resource: this.name,
            db: this._config.db,
            data: data
        };

        if ( data && data.options && data.options.noCache ) {
            env.noCache = true;
        }

        if ( done ) {
            var replyTopic = topic + "." + this.name + "." + uuid.v1();

            var replyTo = {
                channel: resource.config.dataResponseChannel,
                topic: replyTopic
            };

            resource.bus.subscribe({
                channel: replyTo.channel,
                topic: replyTo.topic,
                callback: function( msg, env ) {
                    return done( msg.err, msg.data );
                }
            }).once();

            env.replyTo = replyTo;
        }

        resource.bus.publish( env );
    };

    ResourceFactory.prototype.configure = function( options ) {
        this._config = options;

        if ( options.timestamps ) {
            this.fields({
                ctime: {
                    type: 'number'
                },
                mtime: {
                    type: 'number'
                }
            });
        }

        if ( options.db ) {
            this.db( 'init' );
        }

        return this;
    };

    ResourceFactory.prototype.fullID = function( id ) {

        if ( id.slice( 0, this._key_prefix.length ) === this._key_prefix ) {
            return id;
        }

        return this._key_prefix + id;
    };

    ResourceFactory.prototype.shortID = function( id ) {
        if ( id.slice( 0, this._key_prefix.length ) === this._key_prefix ) {
            return id.slice( this._key_prefix.length );
        }

        return id;
    };

    ResourceFactory.prototype.fields = function( obj ) {
        _.extend( this._schema, obj );
        return this;
    };


    ResourceFactory.prototype.child = function( name, options ) {
        if ( options.embedded === true ) {
            this.embedded = true;
        }
        resource.addRelationship( this.name, name, options );
    };

    ResourceFactory.prototype.parent = function( name, options ) {
        resource.addRelationship( name, this.name, options );
    };

    ResourceFactory.prototype.embedded_in = function( name, options ) {
        options = options || {};
        options.embedded = true;
        this.embedded = true;
        resource.addRelationship( name, this.name, options );
    };

    ResourceFactory.prototype.embeds = function( name, options ) {
        options = options || {};
        options.embedded = true;
        resource.addRelationship( this.name, name, options );
    };

    ResourceFactory.prototype.before = function( key, callback ) {
        this._beforeHooks[ key ] = this._beforeHooks[ key ] || [];
        this._beforeHooks[ key ].push( callback );
        return this;
    };

    ResourceFactory.prototype.after = function( key, callback ) {
        this._afterHooks[ key ] = this._afterHooks[ key ] || [];
        this._afterHooks[ key ].push( callback );
        return this;
    };

    ResourceFactory.prototype.extend = function( instanceMethods, factoryMethods ) {

        if ( ! _.isEmpty( factoryMethods ) ) {
            _.extend( this, factoryMethods );
        }

        if ( ! _.isEmpty( instanceMethods ) ) {
            _.extend( this._instanceMethods, instanceMethods );
        }

        return this;
    };

    ResourceFactory.prototype.update = function( id, data, options, callback ) {
        if ( typeof options === 'function' ) {
            callback = options;
            options = {};
        }

        options.noCache = true;
        callback = callback || noop;

        this.get( id, options, function( err, resource ) {

            if( err ) {
                callback( err );
                return;
            }

            resource.merge( data );

            resource.save( function( err, result ) {

                if ( err ) {
                    callback( err );
                    return;
                }

                callback( '', result );

            });

        });

    };

    ResourceFactory.prototype.save = function( instance, options, callback ) {

    };

    ResourceFactory.prototype.create = function( data, options, callback ) {
        if ( typeof options === 'function' ) {
            callback = options;
            options = {};
        }

        var instance = this.instantiate( data );

        instance.save( function( err, result, meta ) {

            if ( err ) {
                callback( err );
                return;
            }

            if ( callback ) {
                callback( '', instance );
            }

        });

    };

    ResourceFactory.prototype["new"] = function( data ) {
        return this.instantiate( data );
    };

    ResourceFactory.prototype.empty = function() {
        var instance = this.instantiate();

        _.forOwn( instance._properties, function( v, k ) {
            if ( ! v ) {
                instance._properties[ k ] = '';
            }
        });
        
        return instance;
    };

    ResourceFactory.prototype.destroy = function( id, options, callback ) {
        if ( typeof options === 'function' ) {
            callback = options;
            options = {};
        }

        options.noCache = true;
        callback = callback || noop;

        var self = this;

        this.get( id, options, function( err, resource ) {

            if( err ) {
                callback( err );
                return;
            }

            resource.destroy( function( err, result ) {

                if ( err ) {
                    callback( err );
                    return;
                }

                callback( '', result );

            });

        });

        
    };

    ResourceFactory.prototype.instantiate = function( data ) {

        var instance = new resource.instance( this, data );

        if ( this._embedded ) {
            _.merge( instance, resource.mixins.embedded.instance );
        } else {
            _.merge( instance, resource.mixins["default"].instance );
        }

        if ( ! _.isEmpty( this._instanceMethods ) ) {
            _.extend( instance, this._instanceMethods );
        }
        
        return instance;
    };

    _.merge( ResourceFactory.prototype, resource.mixins["default"].factory );

    return ResourceFactory;
};