var _ = require( "lodash" ),
    Monologue = require( "monologue.js" )( _ ),
    revalidator = require( "revalidator" ),
    uuid = require('node-uuid'),
    startsWith = function( subj, str ) {
        if ( subj ) {
            return ( subj.slice( 0, str.length ) === str );
        } else {
            return false;
        }
    };

module.exports = function( resource ) {
    
    var ResourceInstance = function( factory, data ) {
        var self = this;
        self._factory = factory;
        self._name = factory.name;
        self._schema = factory._schema;
        self._timestamps = factory.timestamps;
        self._properties = {};
        self._beforeHooks = factory._beforeHooks;
        self._afterHooks = factory._afterHooks;

        var defaults = {
            id: null
        };

        _.forOwn( self._factory._schema, function( data, key ) {
            defaults[ key ] = null;
        } );

        // Default instance properties
        this.merge( defaults );

        if ( data ) {
            this.merge( data );
        }

        if ( this.id ) {
            this.id = this._factory.shortID( this.id );
        }

    };

    Monologue.mixin( ResourceInstance );

    ResourceInstance.prototype._db = function( topic, data, done ) {
        this._factory.db( topic, data, done );
    };

    ResourceInstance.prototype.merge = function( data ) {

        var self = this,
            new_keys = _.keys( data ),
            existing_keys = _.keys( this._properties ),
            toAdd = _.difference( new_keys, existing_keys );

        if ( data ) {
            _.extend( self._properties, data );
        }

        _.forEach( toAdd, function( key ) {

            self.__defineGetter__( key, function( prop ) {
                return this._properties[ key ];
            });

            self.__defineSetter__( key, function( value ) {
                var old = this._properties[ key ];
                this.emit('field.change.' + key, { name: key, old: old, "new": value } );
                this._properties[ key ] = value;
            });
        });

    };

    var runHooks = function( context, key, done ) {

        var split = key.split( "." ),
            hookArray = split[ 0 ] === 'before' ? '_beforeHooks' : '_afterHooks',
            hookKey = split[ 1 ],
            hooks = context[ hookArray ][ hookKey ],
            runQueue = function( context, queue, done ) {
                var counter = 0,
                    total = queue.length,
                    next = function( err ) {
                        if ( err ) {
                            // Stop execution
                            done( err );
                            return;
                        }

                        if ( counter === ( total - 1 ) ) {
                            done();
                            return;
                        }
                        execute( ++counter );
                    },
                    execute = function( index ) {
                        queue[ index ].call( context, next );
                    };

                execute( 0 );
            };


        if ( _.isArray( hooks ) ) {
            runQueue( context, hooks, done );
        } else {
            done();
        }

    };

    ResourceInstance.prototype.runHooks = runHooks;

    ResourceInstance.prototype.validate = function() {

        var data = {},
            fields = this._schema;

        _.each( this._properties, function( val, key ) {
            if ( ! _.isNull( val ) && ! _.isUndefined( val ) ) {
                data[ key ] = val;
            }
        });

        return revalidator.validate( data, { properties: fields });
    };

    ResourceInstance.prototype.field = function( name, properties ) {
        this._schema[ name ] = properties || {};
    };

    ResourceInstance.prototype.fields = function( obj ) {
        if ( typeof obj === 'object' ) {
            _.cloneDeep( this._schema, obj );
        }
    };

    ResourceInstance.prototype.isNewRecord = function() {
        return ! this._properties.id && ! this._properties._key;
    };

    ResourceInstance.prototype.toJSON = function() {
        return this._properties;
    };

    ResourceInstance.prototype.toString = function() {
        return this._properties;
    };

    return ResourceInstance;
};