var _ = require( "lodash" ),
    inflect = require( "inflect" ),
    uuid = require( "node-uuid" );

// This is a generic callback for processing lists of records.
var list_callback = function( done, context ) {

    return function( err, results ) {
        var self = this;

        if ( err ) {
            return done( err );
        }

        var list = [];

        for( var i = 0, j = results.length; i < j; i++ ) {
            list.push( self.instantiate( results[ i ] ) );
        }

        if ( done ) {
            return process.nextTick(function() {
                return done( '', list );
            });
        }
    }.bind( context );

};

module.exports = {

    "default": {
        instance: {
            save: function( done ) {
                var self = this;

                if ( this._timestamps ) {
                    var now = Date.now();

                    this.mtime = now;

                    if ( this.isNewRecord() ) {
                        this.ctime = now;
                    }
                }

                var id = this.id || uuid.v4();

                if ( this.id ) {
                    this._properties = _.omit( this._properties, 'id' );
                }

                id = this._factory.fullID( id );

                var postSave = function( err, result ) {
                    if ( err ) {
                        done( err );
                        return;
                    }

                    if ( ! self.id ) {
                        self.id = self._factory.shortID( result.id );
                    }

                    self.runHooks( self, 'after.save', function( err ) {
                        done('', self);
                    });

                };

                self.runHooks( this, 'before.save', function( err ) {
                    if ( err ) {
                        done( err );
                        return;
                    }

                    var validation = self.validate();
                    if ( ! validation.valid ) {
                        done( { statusCode: 400, errors: validation.errors, message: "Invalid data" });
                        return;
                    }

                    self._db( "save", {
                        id: id,
                        data: self._properties
                    }, postSave );
                    
                });

            },

            destroy: function( done ) {
                var self = this,
                    id = this._factory.fullID( self.id );

                var postDestroy = function( err, result, meta ) {
                    if ( err ) {
                        done( err );
                        return;
                    }

                    self.runHooks( self, 'after.destroy', function( err ) {
                        done();
                    });

                };

                self.runHooks( this, 'before.destroy', function( err ) {
                    if ( err ) {
                        done( err );
                        return;
                    }

                    self._db( "destroy", {
                        id: id
                    }, postDestroy );
                    
                });

            }
        },
        factory: {
            get: function( id, options, callback ) {
                if ( typeof options === 'function' ) {
                    callback = options;
                    options = {};
                }

                callback = callback || noop;

                var self = this,
                    single = function( err, result ) {
                        if ( err ) {
                            return callback( err );
                        }

                        var instance = self.instantiate( result );

                        if ( callback ) {
                            callback( '', instance );
                        }

                    },
                    multi = function( err, results ) {
                        if ( err ) {
                            return callback( err );
                        }

                        var list = [];
                        
                        for( var i = 0, j = results.length; i < j; i++ ) {
                            list.push( self.instantiate( results[ i ] ) );
                        }

                        if ( callback ) {
                            callback( '', list );
                        }

                    };

                var db_options = {
                    id: _.isArray( id ) ? _.map( id, function( i ) { return self.fullID( i ); } ) : this.fullID( id ),
                    options: options
                };

                var cb = _.isArray( id ) ? multi : single;
                
                this.db( 'get', db_options, cb );
                
            },
            all: function( options, callback ) {
                if ( typeof options === 'function' ) {
                    callback = options;
                    options = {};
                }

                callback = callback || noop;

                var done = list_callback( callback, this );

                var db_options = {
                    options: options
                };

                this.db( "all", db_options, done );

            },

            find: function( keys, options, callback ) {
                if ( ! keys || _.isEmpty( keys ) ) {
                    return process.nextTick( function() {
                        return callback({ statusCode: 404, message: 'No search parameters specified' } );
                    });
                }

                if ( typeof options === 'function' ) {
                    callback = options;
                    options = {};
                }

                callback = callback || noop;

                var done = list_callback( callback, this );

                var db_options = {
                    keys: keys,
                    options: options
                };

                this.db( "find", db_options, done );
            }
        }
    },

    embedded: {
        instance: {
            save: function( done, parent ) {
                var self = this;

                if ( this._timestamps ) {
                    var now = Date.now();

                    this.mtime = now;

                    if ( this.isNewRecord() ) {
                        this.ctime = now;
                    }
                }

                var newRecord = this.id ? false : true;

                this.id = this.id || uuid.v1(),
                this.id = this._factory.fullID( this.id );

                var saveMethod = function( callback ) {
                    var parent_field = self._factory.getParentField(),
                        parentFactory = self._factory.getParent(),
                        child_field = self._factory.name_plural;

                    var parent_insert = function( parentObj, callback ) {
                        if ( ! parentObj[ child_field ] ) {
                            // Child doesn't exist
                            parentObj[ child_field ] = [];
                        }

                        if ( newRecord ) {
                            // Just push into parentObj's child array
                            parentObj[ child_field ].push( self._properties );
                        } else {
                            // Load existing resource and merge
                            var child_index = null,
                                child = _.find( parentObj[ child_field ], function( c, index ) {
                                    if ( self._factory.fullID( c.id ) === self._factory.fullID( self.id ) ) {
                                        child_index = index;
                                        return true;
                                    }
                                    return false;
                                });

                            if ( ! child ) {
                                // Child doesn't exist, push into array
                                parentObj[ child_field ].push( self._properties );
                            } else {
                                // Merge new properties into existing record
                                child = _.extend( child, self._properties );
                                // Replace existing record in parentObj array with merged object
                                parentObj[ child_field ][ child_index ] = child;
                            }
                        }

                        parentObj.save( function( err, p ) {
                            if ( err ) {
                                return callback( err );
                            }

                            callback( null, self._properties );
                        });
                    };

                    if ( parent ) {
                        process.nextTick( function() {
                            parent_insert( parent, callback );
                        });
                    } else {
                        parentFactory.get( self[ parent_field ], { noCache: true }, function( err, parent ) {
                            if ( err ) {
                                return callback( err );
                            }
                            parent_insert( parent, callback );
                        });
                    }

                };

                var postSave = function( err, result ) {
                    if ( err ) {
                        done( err );
                        return;
                    }

                    if ( ! self.id ) {
                        self.id = self._factory.shortID( result.id );
                    }

                    self.runHooks( self, 'after.save', function( err ) {
                        done('', self);
                    });

                };

                self.runHooks( this, 'before.save', function( err ) {
                    if ( err ) {
                        done( err );
                        return;
                    }

                    var validation = self.validate();

                    if ( ! validation.valid ) {
                        done( { statusCode: 400, errors: validation.errors, message: "Invalid data" });
                        return;
                    }

                    saveMethod( postSave );

                });
            },

            destroy: function( done, parent ) {
                var self = this,
                    id = this._factory.fullID( self.id );

                var destroyMethod = function( callback ) {
                    var parent_field = self._factory.getParentField(),
                        parentFactory = self._factory.getParent(),
                        child_field = self._factory.name_plural;

                    var parent_delete = function( parentObj, callback ) {
                        if ( ! parent[ child_field ] ) {
                            // Child doesn't exist
                            return callback( null );
                        }

                        // See if record is here

                        var keep = [],
                            notFound = true;

                        _.each( parent[ child_field ], function( c, index ) {
                            if ( self._factory.fullID( c.id ) === self._factory.fullID( self.id ) ) {
                                notFound = false;
                            } else {
                                keep.push( c );
                            }
                        });


                        if ( notFound ) {
                            // Save a trip to the DB and don't store anything
                            return callback( null );
                        }

                        parent[ child_field ] = keep;

                        parent.save( function( err, p ) {
                            if ( err ) {
                                return callback( err );
                            }

                            callback( null );
                        });
                    };

                    if ( parent ) {
                        process.nextTick( function() {
                            parent_delete( parent, callback );
                        });
                    } else {
                        parentFactory.get( self[ parent_field ], { noCache: true }, function( err, parent ) {
                            if ( err ) {
                                return callback( err );
                            }
                            parent_delete( parent, callback );
                        });
                    }

                };

                var postDestroy = function( err, result, meta ) {
                    if ( err ) {
                        done( err );
                        return;
                    }

                    self.runHooks( self, 'after.destroy', function( err ) {
                        done();
                    });

                };

                self.runHooks( this, 'before.destroy', function( err ) {
                    if ( err ) {
                        done( err );
                        return;
                    }

                    destroyMethod( postDestroy );

                });
            }
        },

        factory: {
            get: function( id, options, done ) {
                if ( typeof options === 'function' ) {
                    done = options;
                    options = {};
                }

                var self = this,
                    parent = self.getParent(),
                    split = id.split( "/" ),
                    getMethod = 'get' + inflect.capitalize( self.name );

                parent[ getMethod ]( split[0], split[1], options, function( err, result ) {
                    if ( err ) {
                        return done( err );
                    }

                    return done( null, result );
                } );
            },

            all: function( parent_id, options, done ) {
                if ( typeof options === 'function' ) {
                    done = options;
                    options = {};
                }

                var self = this,
                    parent = self.getParent(),
                    getMethod = 'get' + inflect.capitalize( self.name_plural );

                parent[ getMethod ]( parent_id, options, function( err, results ) {
                    if ( err ) {
                        return done( err );
                    }

                    return done( null, results );
                });
            }
        }
    }

};