var _ = require( "lodash" ),
    inflect = require( "inflect" ),
    noop = function(){};

var relationships = {
    registry: [],
    addRelationship: function( parent, child, options ) {
        var existing = _.find( relationships.registry, function( r ) {
            return r.parent === parent && r.child === child;
        });

        if ( ! existing ) {
            relationships.registry.push({
                parent: parent,
                child: child,
                options: options
            });
        }
    },
    createRelationship: function( rel, resource ) {
        var parent_name = inflect.singularize( rel.parent ).toLowerCase(),
            child_name = inflect.singularize( rel.child ).toLowerCase(),
            parent = {
                name: parent_name,
                plural: inflect.pluralize( parent_name ),
                factory: resource.get( parent_name )
            },
            child = {
                name: child_name,
                plural: inflect.pluralize( child_name ),
                factory: resource.get( child_name )
            };

        if ( rel.options && rel.options.embedded ) {
            relationships.addEmbeddedMethods( parent, child, rel.options );
        } else {
            relationships.addChildMethods( parent, child, rel.options );
        }
    },

    addChildMethods: function( parent, child, options ) {

        options = options || {};

        var ParentFactory = parent.factory,
            ChildFactory = child.factory,
            cascadeDelete = ( typeof options.cascadeDelete === 'undefined' ) ? true : options.cascadeDelete,
            parent_id_field = ParentFactory.name + "_id",
            child_key_field = child.name + "_ids",

            cname_capitalized = inflect.capitalize( child.name ),
            getMethodPlural = 'get' + inflect.capitalize( child.plural );
            getMethodSingle = 'get' + cname_capitalized,
            updateMethod = 'update' + cname_capitalized,
            destroyMethod = 'destroy' + cname_capitalized,
            createMethod = 'create' + cname_capitalized;

        var parentInstanceMethods = {};

        var methods = {
            fetchChildren: function( parent_id, options, callback ) {
                var query = {};
                query[ parent_id_field ] = parent_id;

                ChildFactory.find( query, options, callback );
            },

            fetchChild: function( parent_id, id, options, callback ) {
                if ( typeof options === 'function' ) {
                    callback = options;
                    options = {};
                }

                callback = callback || noop;

                ChildFactory.get( id, options, function( err, result ) {
                    if ( err ) {
                        callback( err );
                        return;
                    }

                    if ( ParentFactory.fullID( result[ parent_id_field ] ) == ParentFactory.fullID( parent_id ) ) {
                        callback( '', result );
                        return;
                    }

                    callback( { statusCode: 404, message: "not found", notFound: true } );
                });
            },

            createChild: function( parent_id, data, options, callback ) {
                if ( typeof options === 'function' ) {
                    callback = options;
                    options = {};
                }

                data[ parent_id_field ] = ParentFactory.fullID( parent_id );

                ChildFactory.create( data, options, callback );
            },

            updateChild: function( parent_id, id, data, options, callback ) {
                if ( typeof options === 'function' ) {
                    callback = options;
                    options = {};
                }

                options.noCache = true;
                callback = callback || noop;

                ParentFactory[ getMethodSingle ]( parent_id, id, options, function( err, resource ) {

                    if( err ) {
                        callback( err );
                        return;
                    }

                    resource.merge( data );

                    resource.save( function( err, result ) {

                        if ( err ) {
                            callback( err );
                            return;
                        }

                        callback( '', result );

                    });

                });
            },

            destroyChild: function( parent_id, id, options, callback ) {
                if ( typeof options === 'function' ) {
                    callback = options;
                    options = {};
                }
                options.noCache = true;
                callback = callback || noop;

                var self = this;

                ParentFactory[ getMethodSingle ]( parent_id, id, options, function( err, resource ) {

                    if( err ) {
                        callback( err );
                        return;
                    }

                    resource.destroy( function( err, result ) {

                        if ( err ) {
                            callback( err );
                            return;
                        }

                        callback( '', result );

                    });

                });
            }
        };

        
        //                                          \\
        //  Fetches multiple children of a resource \\
        //                                          \\
        parentInstanceMethods[ getMethodPlural ] = function( options, callback ) {
            if ( typeof options === 'function' ) {
                callback = options;
                options = {};
            }

            callback = callback || noop;

            var keys = this[ child_key_field ];
            if ( ! keys ) {
                return callback( '', [] );
            }
            
            methods.fetchChildren( ParentFactory.fullID( this.id ), options, callback );
        };

        ParentFactory[ getMethodPlural ] = function( id, options, callback ) {

            if ( typeof options === 'function' ) {
                callback = options;
                options = {};
            }

            callback = callback || noop;

            ParentFactory.get( id, function( err, result ) {
                if ( err ) {
                    callback( err );
                    return;
                }


                var keys = result[ child_key_field ];

                if ( ! keys ) {
                    return callback( '', [] );
                }

                methods.fetchChildren( ParentFactory.fullID( id ), options, callback );
            });

        };

        /////////////////// End multiple children fetch \\\\\\\\\\\\\\\\\\\\\\

        //                                      \\
        //  Fetches a single child resource     \\
        //                                      \\
        parentInstanceMethods[ getMethodSingle ] = function( id, options, callback ) {
            methods.fetchChild( this.id, id, options, callback );
        };

        ParentFactory[ getMethodSingle ] = function( parent_id, id, options, callback ) {
            methods.fetchChild( parent_id, id, options, callback );
        };

        /////////////////// End single child fetch \\\\\\\\\\\\\\\\\\\\\\

        //                               \\
        //  Creates a child resource     \\
        //                               \\

        parentInstanceMethods[ createMethod ] = function( data, options, callback ) {
            methods.createChild( this.id, data, options, callback );
        };

        ParentFactory[ createMethod ] = function( parent_id, data, options, callback ) {
            methods.createChild( parent_id, data, options, callback );
        };

        /////////////////// End child create \\\\\\\\\\\\\\\\\\\\\\

        //                               \\
        //  Updates a child resource     \\
        //                               \\

        parentInstanceMethods[ updateMethod ] = function( id, data, options, callback ) {
            methods.updateChild( this.id, id, data, options, callback );
        };

        ParentFactory[ updateMethod ] = function( parent_id, id, data, options, callback ) {
            methods.updateChild( parent_id, id, data, options, callback );
        };

        /////////////////// End child update \\\\\\\\\\\\\\\\\\\\\\

        //                               \\
        //  Destroys a child resource    \\
        //                               \\
        parentInstanceMethods[ destroyMethod ] = function( id, options, callback ) {
            methods.destroyChild( this.id, id, options, callback );
        };

        ParentFactory[ destroyMethod ] = function( parent_id, id, options, callback ) {
            console.log(destroyMethod);
            methods.destroyChild( parent_id, id, options, callback );
        };

        /////////////////// End child destroy \\\\\\\\\\\\\\\\\\\\\\

        // Extend instance methods
        ParentFactory.extend( parentInstanceMethods );

        if ( cascadeDelete ) {
            ParentFactory.before( 'destroy', function( done ) {
                ParentFactory[ getMethodPlural ]( this.id, function( err, results ) {
                    if( ! results || ! results.length ) {
                        done();
                        return;
                    }

                    var counter = 0,
                        length = results.length,
                        next = function( err ) {
                            if ( err ) {
                                done( err );
                                return;
                            }

                            if ( counter == ( length - 1 ) ) {
                                done();
                            } else {
                                destroy( ++counter );
                            }
                        },
                        destroy = function( index ) {

                            results[ index ].destroy( next );

                        };

                    destroy( 0 );
                });

            });
        }

        // Handle adding and removing child id to parent object
        ChildFactory.after( 'save', function( done ) {
            
            var parent_id = this[ parent_id_field ],
                fullID = this._factory.fullID( this.id );

            if ( ! parent_id ) {
                return done();
            }

            ParentFactory.get( parent_id, { noCache: true }, function( err, parent ) {

                if ( err ) {
                    return done({ statusCode: 404, msg: 'Parent not found' });
                }

                var store = false,
                    children = parent[ child_key_field ];

                if ( ! children ) {
                    parent[ child_key_field ] = [];
                }

                if ( ! _.contains( children, fullID ) ) {
                    store = true;
                    parent[ child_key_field ].push( fullID );
                }

                if ( ! store ) {
                    // Child ID already exists in parent
                    return done();
                }

                parent.save( function( err, p ) {
                    if ( err ) {
                        return done({ statusCode: 404, msg: 'Parent could not be saved' });
                    }
                    done();
                });

            });
            
        });

        ChildFactory.after( 'destroy', function( done ) {
            
            var parent_id = this[ parent_id_field ],
                fullID = this._factory.fullID( this.id );

            if ( ! parent_id ) {
                return done();
            }

            ParentFactory.get( parent_id, { noCache: true }, function( err, parent ) {

                if ( err ) {
                    return done({ statusCode: 404, msg: 'Parent not found' });
                }

                var store = false,
                    children = parent[ child_key_field ];

                if ( ! children || ! _.contains( children, fullID ) ) {
                    return done();
                }

                parent[ child_key_field ] = _.without( children, fullID );

                parent.save( function( err, p ) {
                    if ( err ) {
                        return done({ statusCode: 404, msg: 'Parent could not be saved' });
                    }
                    done();
                });

            });
            
        });

        return this;
    },

    addEmbeddedMethods: function( parent, child, options ) {

        options = options || {};

        var ParentFactory = parent.factory,
            ChildFactory = child.factory,
            cascadeDelete = ( typeof options.cascadeDelete === 'undefined' ) ? true : options.cascadeDelete;


        var parent_id_field = parent.name + "_id";
            addFields = {};
            addFields[ parent_id_field ] = {
                type: 'string',
                required: 'true'
            };

        ChildFactory.fields( addFields );

        ChildFactory.getParentField = function() {
            return parent_id_field;
        };

        ChildFactory.getParent = function() {
            return ParentFactory;
        };

        var cname_capitalized = inflect.capitalize( child.name ),
            getMethodSingle = 'get' + cname_capitalized,
            getMethodPlural = 'get' + inflect.capitalize( child.plural ),
            updateMethod = 'update' + cname_capitalized,
            destroyMethod = 'destroy' + cname_capitalized,
            createMethod = 'create' + cname_capitalized;

        var parentInstanceMethods = {},
            methods = {
                fetchChildren: function( children, callback ) {
                    if ( children.length ) {
                        children = _.map( children, function( c ) {
                            return ChildFactory.instantiate( c );
                        });
                    }

                    callback( null, children );
                },
                fetchChild: function( id, children, callback ) {
                    var child = _.find( children, function( c ) {
                        return ChildFactory.fullID( c.id ) === ChildFactory.fullID( id );
                    });

                    if ( ! child ) {
                        return callback( { statusCode: 404, message: 'not found'} );
                    }

                    return callback( null, child );
                },
                createChild: function( parent_id, data, options, callback ) {
                    if ( typeof options === 'function' ) {
                        callback = options;
                        options = {};
                    }

                    ChildFactory.create( data, options, callback );
                },
                updateChild: function() {
                    if ( typeof options === 'function' ) {
                        callback = options;
                        options = {};
                    }

                    ChildFactory.update( id, data, options, callback );
                },
                destroyChild: function() {

                }
            };

        //                                          \\
        //  Fetches multiple children of a resource \\
        //                                          \\
        parentInstanceMethods[ getMethodPlural ] = function( options, callback ) {
            if ( typeof options === 'function' ) {
                callback = options;
                options = {};
            }

            callback = callback || noop;
            var field = child.plural;

            var children = this[ field ] || [];
            methods.fetchChildren( children, callback );
        };

        ParentFactory[ getMethodPlural ] = function( id, options, callback ) {

            if ( typeof options === 'function' ) {
                callback = options;
                options = {};
            }

            callback = callback || noop;

            var self = this,
                field = child.plural;

            self.get( id, options, function( err, result ) {
                if ( err ) {
                    return callback( err );
                }

                var children = result[ field ] || [];

                var cb = function( err, results ) {
                    callback( err, results, result );
                };

                methods.fetchChildren( children, cb );

            });

        };

        /////////////////// End fetch children \\\\\\\\\\\\\\\\\\\\\\

        //                                          \\
        //  Fetches single child of a resource      \\
        //                                          \\
        parentInstanceMethods[ getMethodSingle ] = function( id, options, callback ) {
            if ( typeof options === 'function' ) {
                callback = options;
                options = {};
            }

            var self = this;
            self[ getMethodPlural ]( function( err, results ) {
                methods.fetchChild( id, results, callback );
            });
        };

        ParentFactory[ getMethodSingle ] = function( parent_id, id, options, callback ) {
            if ( typeof options === 'function' ) {
                callback = options;
                options = {};
            }

            callback = callback || noop;
            var self = this;

            self[ getMethodPlural ]( parent_id, options, function( err, children ) {
                if ( err ) {
                    return callback( err );
                }

                methods.fetchChild( id, children, callback );
            });

        };

        /////////////////// End child fetch \\\\\\\\\\\\\\\\\\\\\\

        //                                          \\
        //  Creates a single child of a resource    \\
        //                                          \\
        parentInstanceMethods[ createMethod ] = function( data, options, callback ) {
            if ( typeof options === 'function' ) {
                callback = options;
                options = {};
            }

            var child = ChildFactory.instantiate( data );
            child[ parent_id_field ] = this._factory.fullID( this.id );
            child.save( callback, this );
        };

        ParentFactory[ createMethod ] = function( parent_id, data, options, callback ) {
            methods.createChild( parent_id, data, options, callback );
        };

        /////////////////// End child create \\\\\\\\\\\\\\\\\\\\\\
        
        //                                          \\
        //  Updates a child of a resource    \\
        //                                          \\
        parentInstanceMethods[ updateMethod ] = function( id, data, options, callback ) {
            if ( typeof options === 'function' ) {
                callback = options;
                options = {};
            }

            var self = this;

            self[ getMethodSingle ]( id, function( err, child ) {
                if ( err ) {
                    return callback( err );
                }

                child.merge( data );

                child[ parent_id_field ] = self._factory.fullID( self.id );

                child.save( callback, self );
            });

        };

        ParentFactory[ updateMethod ] = function( id, data, options, callback ) {
            methods.updateChild( id, data, options, callback );
        };

        /////////////////// End child update \\\\\\\\\\\\\\\\\\\\\\
        
        //                                          \\
        //  Deletes single child of a resource    \\
        //                                          \\
        parentInstanceMethods[ destroyMethod ] = function( id, options, callback ) {
            if ( typeof options === 'function' ) {
                callback = options;
                options = {};
            }

            var self = this;

            self[ getMethodSingle ]( id, function( err, child ) {
                if ( err ) {
                    return callback( err );
                }

                child.destroy( callback, self );
            });
        };

        ParentFactory[ destroyMethod ] = function( parent_id, data, options, callback ) {
            methods.destroyChild( parent_id, data, options, callback );
        };

        /////////////////// End child delete \\\\\\\\\\\\\\\\\\\\\\

        // Extend instance methods
        ParentFactory.extend( parentInstanceMethods );

        return this;
    }
};
module.exports = relationships;