require("./_setup.js")( global );

describe( "ResourceFactory", function() {

    var dbMock = null,
        Article,
        Comment;

    before( function( done ) {
        resource.reset();
        resource.Client.off();
        dbMock = new DbMock();

        Article = resource.define( 'article', function() {
            this.configure({
                bucket: 'articles',
                client: {
                    hosts: couchbase_config.hosts,
                    user: couchbase_config.user,
                    password: couchbase_config.password,
                    bucket: 'articles'
                },
                timestamps: true
            });

            this.fields({
                name: {
                    type: 'string',
                    required: true
                },
                slug: {
                    type: 'string',
                    required: true
                },
                description: {
                    type: 'string'
                }
            });

            this.child( 'comment' );

        });

        Article.extend(
            {
                getData: function() {
                    return "GETTING DATA AGAIN";
                }
            },
            {
                getByTopic: function() {
                    return "GETTING BY TOPIC";
                }
            }
        );

        Comment = resource.define( 'comment', function() {
            this.configure({
                bucket: 'comments',
                client: {
                    hosts: couchbase_config.hosts,
                    user: couchbase_config.user,
                    password: couchbase_config.password,
                    bucket: 'comments'
                },
                timestamps: true
            });

            this.fields({
                text: {
                    type: 'string'
                }
            });
        });


        installer.run( function() {
            done();
        });
    });

    after( function( done ) {
        resource.reset();
        dbMock.clear();
        installer.run( function() {
            done();
        });
    });

    describe( "all", function() {
        it( "should return a list of articles", function( done ) {
            Article.all( function( err, results ) {
                expect( results.length ).to.be.greaterThan( 0 );
                done();
            });

        });
    });

    describe( "get", function() {
        it( "should fetch a single article", function( done ) {
            var id = '0015',
                change_counter = 0;

            Article.get( id, function( err, p ) {
                p.on('field.change.slug', function( msg, env ) {
                    change_counter++;
                    expect( msg["new"] ).to.equal( 'some-other-slug' );
                });

                p.slug = 'some-other-slug';
                expect( p.id ).to.equal( id );
                expect( change_counter ).to.equal( 1 );

                p.getData();

                done();
            });

        });

        it( "should fetch multiple articles", function( done ) {
            var id = ['0015', '0016'],
                change_counter = 0;

            Article.get( id, function( err, articles ) {


                expect( articles.length ).to.equal( 2 );

                expect( articles[0].id ).to.equal( '0015' );
                expect( articles[1].id ).to.equal( '0016' );

                done();
            });

        });
    });

    describe( "child methods", function() {
        it( "should add a method that retrieves all child resources", function( done ) {
            Article.comments( '0012', function( err, results ) {
                expect( results.length ).to.be.greaterThan( 0 );
                done();
            });
        });

        it( "should add a method that retrieves a single child resource", function( done ) {
            Article.comment( '0012', '0111', function( err, result ) {
                expect( result.id ).to.equal( '0111' );
                done();
            });
        });

    });

});

