require("./_setup.js")( global );

var postal = require( 'postal' )();

describe( "ResourceInstance", function() {

    var Presentation = resource.define( 'presentation', function( self ) {
        this.configure({
            bucket: 'presentations',
            client: {
                encodeUri: true,
                host: riak_config.host,
                port: riak_config.port
            }
        });

        this.fields({
            title: {
                type: 'string',
                required: true
            },
            slug: {
                type: 'string',
                required: true
            },
            description: {
                type: 'string'
            }
        });

    });

    describe( "when initializing an instance", function() {

        it( "should merge in data correctly", function() {

            var data = {
                id: 'presentation/0013',
                title: 'Contextual jQuery',
                slug: 'contextual-jquery',
                author: 'Eli Perelman',
                slideCount: 26,
                exerciseCount: 15,
                level: 'Intermediate',
                thumbnail: '/thumbnails/thumbnail-3.jpg',
                description: '<p>Thinking out of the box with jQuery</p>',
                topic_id: '0002'
            };

            var expected = _.clone( data );

            expected._key = 'presentation/0013';
            expected.id = '0013';

            var p = Presentation.instantiate( data );

            expect( p.id ).to.equal( '0013' );
            expect( p._key ).to.equal( 'presentation/0013' );
            expect( p.toJSON() ).to.eql( expected );

        });

        it( "should create fill in id when only key is present", function() {
            var data = {
                _key: 'presentation/0013',
                title: 'Contextual jQuery',
                slug: 'contextual-jquery',
                author: 'Eli Perelman',
                slideCount: 26,
                exerciseCount: 15,
                level: 'Intermediate',
                thumbnail: '/thumbnails/thumbnail-3.jpg',
                description: '<p>Thinking out of the box with jQuery</p>',
                topic_id: '0002'
            };

            var p = Presentation.instantiate( data );

            var expected = _.clone( data );

            expected._key = 'presentation/0013';
            expected.id = '0013';

            expect( p.id ).to.equal( '0013' );
            expect( p._key ).to.equal( 'presentation/0013' );
            expect( p.toJSON() ).to.eql( expected );
        });

        it( "should create fill in key when only abbreviated id is present", function() {
            var data = {
                _key: 'presentation/0013',
                title: 'Contextual jQuery',
                slug: 'contextual-jquery',
                author: 'Eli Perelman',
                slideCount: 26,
                exerciseCount: 15,
                level: 'Intermediate',
                thumbnail: '/thumbnails/thumbnail-3.jpg',
                description: '<p>Thinking out of the box with jQuery</p>',
                topic_id: '0002'
            };

            var p = Presentation.instantiate( data );

            var expected = _.clone( data );

            expected._key = 'presentation/0013';
            expected.id = '0013';

            expect( p.id ).to.equal( '0013' );
            expect( p._key ).to.equal( 'presentation/0013' );
            expect( p.toJSON() ).to.eql( expected );
        });

        it( "should create not fill in id or key if neither are present", function() {
            var data = {
                title: 'Contextual jQuery',
                slug: 'contextual-jquery',
                author: 'Eli Perelman',
                slideCount: 26,
                exerciseCount: 15,
                level: 'Intermediate',
                thumbnail: '/thumbnails/thumbnail-3.jpg',
                description: '<p>Thinking out of the box with jQuery</p>',
                topic_id: '0002'
            };

            var p = Presentation.instantiate( data );

            var expected = _.clone( data );

            expected.id = null;
            expected._key = null;

            expect( p.id ).to.be( null );
            expect( p._key ).to.be( null );
            expect( p.toJSON() ).to.eql( expected );
        });

    });

    describe( "save", function() {

        it( "should persist an instance to the database", function( done ) {

            var pres = Presentation.new({
                title: 'My Presentation',
                slug: 'my-presentation',
                description: 'Some stuff i want to talk about'
            });

            pres.save( function( err, p ) {
                expect( p.id ).to.be.ok();
                expect( p._key ).to.equal( 'presentation/' + p.id );
                done();
            });

        });

        it( "should fail if data is invalid", function( done ) {

            var pres = Presentation.new({
                slug: 'my-presentation',
                description: 'Some stuff i want to talk about'
            });

            pres.save( function( err, p ) {
                expect( err.statusCode ).to.equal( 400 );
                expect( err.message ).to.equal( "Invalid data" );
                done();
            });

        });

    });

    describe( "save hooks", function() {

        var SomeResource = resource.define( 'someresource', function() {

            this.configure({
                bucket: 'someresources',
                client: {
                    encodeUri: true,
                    host: riak_config.host,
                    port: riak_config.port
                }
            });

            this.fields({
                title: {
                    type: 'string',
                    required: true
                },
                slug: {
                    type: 'string',
                    required: true
                },
                description: {
                    type: 'string'
                }
            });

            this.before( 'save', function( done ) {
                postal.publish({
                    channel: 'hooks',
                    topic: 'before.save',
                    data: {
                        msg: 'before 1'
                    }
                });
                done();
            });

            this.before( 'save', function( done ) {
                
                postal.publish({
                    channel: 'hooks',
                    topic: 'before.save',
                    data: {
                        msg: 'before 2'
                    }
                });
                done();
            });

            this.after( 'save', function( done ) {
                postal.publish({
                    channel: 'hooks',
                    topic: 'after.save',
                    data: {
                        msg: 'after 1'
                    }
                });
                done();
            });

            this.after( 'save', function( done ) {
                postal.publish({
                    channel: 'hooks',
                    topic: 'after.save',
                    data: {
                        msg: 'after 2'
                    }
                });
                done();
            });

        });

        it( 'should run before save hooks if present', function( done ) {

            var r = SomeResource.new({
                title: 'My Presentation',
                slug: 'my-presentation',
                description: 'Some stuff i want to talk about'
            });

            var before = [];
            postal.subscribe({
                channel: "hooks",
                topic: "before.save",
                callback: function( data, env ) {
                    before.push( data.msg );
                }
            });

            r.save( function( err, p ) {
                expect( p.id ).to.be.ok();
                expect( p._key ).to.equal( 'someresource/' + p.id );
                expect( before[ 0 ] ).to.equal( 'before 1' );
                expect( before[ 1 ] ).to.equal( 'before 2' );
                done();
            });

        });

        it( 'should run after save hooks if present', function( done ) {

            var r = SomeResource.new({
                title: 'My Presentation',
                slug: 'my-presentation',
                description: 'Some stuff i want to talk about'
            });

            var after = [];
            postal.subscribe({
                channel: "hooks",
                topic: "after.save",
                callback: function( data, env ) {
                    after.push( data.msg );
                }
            });

            r.save( function( err, p ) {
                expect( p.id ).to.be.ok();
                expect( p._key ).to.equal( 'someresource/' + p.id );
                expect( after[ 0 ] ).to.equal( 'after 1' );
                expect( after[ 1 ] ).to.equal( 'after 2' );
                done();
            });

        });

        it( 'should prevent saving if a before hook passes back an error', function( done ) {

            SomeResource.before( 'save', function( done ) {
                postal.publish({
                    channel: 'hooks',
                    topic: 'before.save',
                    data: {
                        msg: 'before 3'
                    }
                });
                done({
                    statusCode: 400,
                    msg: 'There was an error'
                });
            });

            var r = SomeResource.new({
                title: 'My Presentation',
                slug: 'my-presentation',
                description: 'Some stuff i want to talk about'
            });

            var before = [];
            postal.subscribe({
                channel: "hooks",
                topic: "before.save",
                callback: function( data, env ) {
                    before.push( data.msg );
                }
            });

            r.save( function( err, p ) {
                expect( err.statusCode ).to.equal( 400 );
                expect( err.msg ).to.equal( 'There was an error' );
                
                expect( before[ 0 ] ).to.equal( 'before 1' );
                expect( before[ 1 ] ).to.equal( 'before 2' );
                expect( before[ 2 ] ).to.equal( 'before 3' );
                done();
            });

        });

    });

    describe( "timestamps", function() {

        var SomeResource = resource.define( 'someresource', function() {

                this.configure({
                    bucket: 'someresources',
                    client: {
                        encodeUri: true,
                        host: riak_config.host,
                        port: riak_config.port
                    },
                    timestamps: true
                });

                this.fields({
                    title: {
                        type: 'string',
                        required: true
                    },
                    slug: {
                        type: 'string',
                        required: true
                    },
                    description: {
                        type: 'string'
                    }
                });

            });

        it( "should save created time and modifed time to object if specified", function( done ) {
            expect( SomeResource._schema ).to.have.property( 'mtime' );
            expect( SomeResource._schema ).to.have.property( 'ctime' );

            var r = SomeResource.new({
                title: 'New Resource',
                slug: 'new-resource',
                description: 'some description'
            });

            expect( r.isNewRecord() ).to.be.ok();

            r.save( function( err, result ) {
                expect( result.mtime ).to.be.ok();
                expect( result.ctime ).to.be.ok();
                expect( result.isNewRecord() ).to.not.be.ok();
                done();
            });

        });
    });

});

